#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import subscription_manager.injection as inj

from subscription_manager.cli_command.cli import CliCommand
from subscription_manager.i18n import ugettext as _


class AutohealCommand(CliCommand):
    def __init__(self):
        self.uuid = inj.require(inj.IDENTITY).uuid

        shortdesc = _("Set if subscriptions are attached on a schedule (default of daily)")
        self._org_help_text = _("specify whether to enable or disable auto-attaching of subscriptions")
        super(AutohealCommand, self).__init__("auto-attach", shortdesc, False)

        self.parser.add_argument(
            "--enable",
            dest="enable",
            action="store_true",
            help=_("try to attach subscriptions for uncovered products each check-in"),
        )
        self.parser.add_argument(
            "--disable",
            dest="disable",
            action="store_true",
            help=_("do not try to automatically attach subscriptions each check-in"),
        )
        self.parser.add_argument(
            "--show",
            dest="show",
            action="store_true",
            help=_("show the current auto-attach preference"),
        )

    def _toggle(self, autoheal):
        self.cp.updateConsumer(self.uuid, autoheal=autoheal)
        self._show(autoheal)

    def _validate_options(self):
        if not self.uuid:
            self.assert_should_be_registered()

    def _show(self, autoheal):
        if autoheal:
            print(_("Auto-attach preference: enabled"))
        else:
            print(_("Auto-attach preference: disabled"))

    def _do_command(self):
        self._validate_options()

        if not self.options.enable and not self.options.disable:
            self._show(self.cp.getConsumer(self.uuid)["autoheal"])
        else:
            self._toggle(self.options.enable or False)



# Copyright (c) 2013 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public
# License as published by the Free Software Foundation; either version
# 2 of the License (GPLv2) or (at your option) any later version.
# There is NO WARRANTY for this software, express or implied,
# including the implied warranties of MERCHANTABILITY,
# NON-INFRINGEMENT, or FITNESS FOR A PARTICULAR PURPOSE. You should
# have received a copy of GPLv2 along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
import errno
import socket
import unittest

from subscription_manager.exceptions import ExceptionMapper
from subscription_manager.certdirectory import DEFAULT_PRODUCT_CERT_DIR
from rhsm.connection import (
    RestlibException,
    BadCertificateException,
    ProxyException,
    UnknownContentException,
    ConnectionOSErrorException,
)
from rhsm.https import httplib, ssl
from rhsm.certificate2 import CertificateLoadingError


class MyRuntimeErrorBase(RuntimeError):
    def __init__(self, message):
        RuntimeError.__init__(self, message)


class MyRuntimeError(MyRuntimeErrorBase):
    def __init__(self, message):
        MyRuntimeErrorBase.__init__(self, message)


class OldStyleClass:
    def __init__(self):
        pass


class TestExceptionMapper(unittest.TestCase):
    def test_single_mapped_exception(self):
        expected_message = "Single Exception Message"
        mapper = ExceptionMapper()
        mapper.message_map[RuntimeError] = (expected_message, mapper.format_using_template)

        err = RuntimeError("Testing")
        self.assertEqual(expected_message, mapper.get_message(err))

    def test_subclass_mapped_by_base_class(self):
        expected_message = "Single Exception Message"
        mapper = ExceptionMapper()
        mapper.message_map[RuntimeError] = (expected_message, mapper.format_using_template)

        err = MyRuntimeError("Testing base class")
        self.assertEqual(expected_message, mapper.get_message(err))

    def test_subclass_preferred_over_base_class(self):
        expected_message = "Subclass Exception Message"
        mapper = ExceptionMapper()
        mapper.message_map[RuntimeError] = ("RuntimeError message", mapper.format_using_template)
        mapper.message_map[MyRuntimeErrorBase] = ("MyRuntimeErrorBase message", mapper.format_using_template)
        mapper.message_map[MyRuntimeError] = (expected_message, mapper.format_using_template)

        err = MyRuntimeError("Logged Only")
        self.assertEqual(expected_message, mapper.get_message(err))

    def test_can_map_middle_sub_class(self):
        expected_message = "MyRuntimeErrorBase message"
        mapper = ExceptionMapper()
        mapper.message_map[RuntimeError] = ("RuntimeError message", mapper.format_using_template)
        mapper.message_map[MyRuntimeErrorBase] = (expected_message, mapper.format_using_template)
        mapper.message_map[MyRuntimeError] = ("MyRuntimeError message", mapper.format_using_template)

        err = MyRuntimeErrorBase("Logged Only")
        self.assertEqual(expected_message, mapper.get_message(err))

    def test_search_for_base_class_with_gaps(self):
        mapper = ExceptionMapper()
        expected_message = "RuntimeError message"

        mapper.message_map[RuntimeError] = (expected_message, mapper.format_using_template)
        err = MyRuntimeError("Logged Only")
        self.assertEqual(expected_message, mapper.get_message(err))

    def test_restlib_exception_uses_custom_message(self):
        expected_message = "Expected MESSAGE"
        mapper = ExceptionMapper()

        err = RestlibException(404, expected_message)
        self.assertEqual(f"{expected_message} (HTTP error code 404: Not Found)", mapper.get_message(err))

    def test_return_str_when_no_mapped_exception(self):
        expected_message = "Expected message"
        mapper = ExceptionMapper()

        err = RuntimeError(expected_message)
        self.assertEqual(expected_message, mapper.get_message(err))

    def test_can_support_old_style_classes(self):
        expected_message = "Old style class"
        mapper = ExceptionMapper()
        mapper.message_map[OldStyleClass] = (expected_message, mapper.format_using_template)

        err = OldStyleClass()
        self.assertEqual(expected_message, mapper.get_message(err))

    def test_bad_certificate_exception(self):
        expected_message = "Expected MESSAGE"
        mapper = ExceptionMapper()

        sslerr = ssl.SSLError(5, expected_message)
        err = BadCertificateException("foo.pem", sslerr)
        self.assertEqual(f"Bad CA certificate: foo.pem: {expected_message}", mapper.get_message(err))

    def test_connectionerror(self):
        expected_message = "Expected MESSAGE"
        expected_errno = errno.ECONNREFUSED
        mapper = ExceptionMapper()

        err = ConnectionRefusedError(expected_errno, expected_message)
        self.assertEqual(
            f"Connection error: {expected_message} (error code {expected_errno})",
            mapper.get_message(err),
        )

    def test_socket_gaierror(self):
        expected_message = "Expected MESSAGE"
        expected_errno = socket.EAI_NONAME
        mapper = ExceptionMapper()

        err = socket.gaierror(expected_errno, expected_message)
        self.assertEqual(
            f"Network error: {expected_message} (error code {expected_errno})",
            mapper.get_message(err),
        )

    def test_proxyexception_with_exception_oserror(self):
        expected_message = "Expected MESSAGE"
        expected_errno = errno.ECONNREFUSED
        expected_hostname = "hostname"
        expected_port = 1234
        mapper = ExceptionMapper()

        oserr = ConnectionRefusedError(expected_errno, expected_message)
        err = ProxyException(expected_hostname, expected_port, oserr)
        self.assertEqual(
            f"Proxy error: unable to connect to {expected_hostname}:{expected_port}: "
            f"{expected_message} (error code {expected_errno})",
            mapper.get_message(err),
        )

    def test_proxyexception_with_exception_oserror_no_errno(self):
        expected_message = "Expected MESSAGE"
        expected_hostname = "hostname"
        expected_port = 1234
        mapper = ExceptionMapper()

        oserr = OSError(expected_message)
        err = ProxyException(expected_hostname, expected_port, oserr)
        self.assertEqual(
            f"Proxy error: unable to connect to {expected_hostname}:{expected_port}: {expected_message}",
            mapper.get_message(err),
        )

    def test_proxyexception_with_exception_non_oserror(self):
        expected_message = "Expected MESSAGE"
        expected_hostname = "hostname"
        expected_port = 1234
        mapper = ExceptionMapper()

        genericerr = Exception(expected_message)
        err = ProxyException(expected_hostname, expected_port, genericerr)
        self.assertEqual(
            f"Proxy error: unable to connect to {expected_hostname}:{expected_port}: {expected_message}",
            mapper.get_message(err),
        )

    def test_proxyexception_without_exception(self):
        expected_hostname = "hostname"
        expected_port = 1234
        mapper = ExceptionMapper()

        err = ProxyException(expected_hostname, expected_port)
        self.assertEqual(
            f"Proxy error: unable to connect to {expected_hostname}:{expected_port}",
            mapper.get_message(err),
        )

    def test_unknowncontentexception_with_content(self):
        expected_http_code = 404
        expected_http_string = httplib.responses[expected_http_code]
        expected_content_type = "text/plain"
        expected_content_original = "\033[92mExpected GREEN MESSAGE\033[0m"
        expected_content_escaped = "<27>[92mExpected GREEN MESSAGE<27>[0m"
        mapper = ExceptionMapper()

        err = UnknownContentException(expected_http_code, expected_content_type, expected_content_original)
        self.assertEqual(
            f"Unknown server reply (HTTP error code {expected_http_code}: "
            f"{expected_http_string}):\n{expected_content_escaped}",
            mapper.get_message(err),
        )

    def test_unknowncontentexception_without_content(self):
        expected_http_code = 404
        expected_http_string = httplib.responses[expected_http_code]
        mapper = ExceptionMapper()

        err = UnknownContentException(expected_http_code)
        self.assertEqual(
            f"Unknown server reply (HTTP error code {expected_http_code}: {expected_http_string})",
            mapper.get_message(err),
        )

    def test_certificateloadingerror_product_certificate(self):
        expected_library = "X509"
        expected_reason = "Expected MESSAGE"
        expected_path = f"{DEFAULT_PRODUCT_CERT_DIR}/foo.pem"
        mapper = ExceptionMapper()

        err = CertificateLoadingError(expected_library, expected_reason, path=expected_path)
        self.assertEqual(
            f"Bad product certificate: {expected_path}: [{expected_library}] {expected_reason}",
            mapper.get_message(err),
        )

    def test_certificateloadingerror_other_certificate(self):
        expected_library = "X509"
        expected_reason = "Expected MESSAGE"
        expected_path = "/tmp/foo.pem"
        mapper = ExceptionMapper()

        err = CertificateLoadingError(expected_library, expected_reason, path=expected_path)
        self.assertEqual(
            f"Bad certificate: {expected_path}: [{expected_library}] {expected_reason}",
            mapper.get_message(err),
        )

    def test_certificateloadingerror_pem_data(self):
        expected_library = "X509"
        expected_reason = "Expected MESSAGE"
        expected_data = "DOESTHISLOOKLIKEACERT?"
        mapper = ExceptionMapper()

        err = CertificateLoadingError(expected_library, expected_reason, pem=expected_data)
        self.assertEqual(
            f"Bad certificate: [{expected_library}] {expected_reason}\n{expected_data}",
            mapper.get_message(err),
        )

    def test_certificateloadingerror_unknown(self):
        expected_library = "X509"
        expected_reason = "Expected MESSAGE"
        mapper = ExceptionMapper()

        err = CertificateLoadingError(expected_library, expected_reason)
        self.assertEqual(
            f"Bad certificate: [{expected_library}] {expected_reason}",
            mapper.get_message(err),
        )

    def test_connectionoserror(self):
        expected_message = "Expected MESSAGE"
        expected_hostname = "hostname"
        expected_port = 1234
        expected_handler = "/foo"
        mapper = ExceptionMapper()

        genericerr = OSError(expected_message)
        err = ConnectionOSErrorException(expected_hostname, expected_port, expected_handler, genericerr)
        self.assertEqual(
            f"Unable to reach the server at {expected_hostname}:{expected_port}{expected_handler}: "
            f"{expected_message}",
            mapper.get_message(err),
        )



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import logging

from subscription_manager.cli_command.cli import CliCommand
from subscription_manager.i18n import ugettext as _

log = logging.getLogger(__name__)


class VersionCommand(CliCommand):
    def __init__(self):
        shortdesc = _("Print version information")

        super(VersionCommand, self).__init__("version", shortdesc, False)

    def _do_command(self):
        self.log_server_version()
        print(_("server type: {type}").format(type=self.server_versions["server-type"]))
        print(
            _("subscription management server: {version}").format(version=self.server_versions["candlepin"])
        )
        print(
            _("subscription management rules: {version}").format(
                version=self.server_versions["rules-version"]
            )
        )
        print("subscription-manager: {version}".format(version=self.client_versions["subscription-manager"]))



# Copyright (c) 2017 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
from unittest import mock
import json

from rhsmlib.dbus.objects import EntitlementDBusObject

from test.rhsmlib.base import DBusServerStubProvider


class TestEntitlementDBusObject(DBusServerStubProvider):
    dbus_class = EntitlementDBusObject
    dbus_class_kwargs = {}

    def test_GetStatus(self):
        get_status_patch = mock.patch(
            "rhsmlib.services.entitlement.EntitlementService.get_status",
            name="get_status",
        )
        self.patches["get_status_patch"] = get_status_patch.start()
        self.addCleanup(get_status_patch.stop)

        status = {"status": "Unknown", "reasons": {}, "valid": False}
        self.patches["get_status_patch"].return_value = status

        expected = json.dumps(status)
        result = self.obj.GetStatus.__wrapped__(self.obj, "", self.LOCALE)
        self.assertEqual(expected, result)

    def test_RemoveEntitlementsBySerials(self):
        remove_entitlements_by_serials_patch = mock.patch(
            "rhsmlib.services.entitlement.EntitlementService.remove_entitlements_by_serials",
            name="remove_entitlements_by_serials",
        )
        self.patches["remove_entitlements_by_serials"] = remove_entitlements_by_serials_patch.start()
        self.addCleanup(remove_entitlements_by_serials_patch.stop)

        removed_nonremoved = (["123"], [])
        self.patches["remove_entitlements_by_serials"].return_value = removed_nonremoved

        expected = json.dumps(removed_nonremoved[0])
        result = self.obj.RemoveEntitlementsBySerials.__wrapped__(self.obj, ["123"], {}, self.LOCALE)
        self.assertEqual(expected, result)

    def test_RemoveEntitlementsBySerials__multiple(self):
        remove_entitlements_by_serials_patch = mock.patch(
            "rhsmlib.services.entitlement.EntitlementService.remove_entitlements_by_serials",
            name="remove_entitlements_by_serials",
        )
        self.patches["remove_entitlements_by_serials"] = remove_entitlements_by_serials_patch.start()
        self.addCleanup(remove_entitlements_by_serials_patch.stop)

        removed_nonremoved = (["123", "456"], [])
        self.patches["remove_entitlements_by_serials"].return_value = removed_nonremoved

        expected = json.dumps(removed_nonremoved[0])
        result = self.obj.RemoveEntitlementsBySerials.__wrapped__(self.obj, ["123", "456"], {}, self.LOCALE)
        self.assertEqual(expected, result)

    def test_RemoveEntitlementsBySerials__good_and_bad(self):
        remove_entitlements_by_serials_patch = mock.patch(
            "rhsmlib.services.entitlement.EntitlementService.remove_entitlements_by_serials",
            name="remove_entitlements_by_serials",
        )
        self.patches["remove_entitlements_by_serials"] = remove_entitlements_by_serials_patch.start()
        self.addCleanup(remove_entitlements_by_serials_patch.stop)

        removed_nonremoved = (["123"], ["456"])
        self.patches["remove_entitlements_by_serials"].return_value = removed_nonremoved

        expected = json.dumps(removed_nonremoved[0])
        result = self.obj.RemoveEntitlementsBySerials.__wrapped__(self.obj, ["123", "789"], {}, self.LOCALE)
        self.assertEqual(expected, result)

    def test_RemoveEntitlementsByPoolIds(self):
        remove_entitlements_by_pool_ids_patch = mock.patch(
            "rhsmlib.services.entitlement.EntitlementService.remove_entitlements_by_pool_ids",
            name="remove_entitlements_by_pool_ids",
        )
        self.patches["remove_entitlements_by_pool_ids"] = remove_entitlements_by_pool_ids_patch.start()
        self.addCleanup(remove_entitlements_by_pool_ids_patch.stop)

        removed_nonremoved_serials = (["123"], [], ["456"])
        self.patches["remove_entitlements_by_pool_ids"].return_value = removed_nonremoved_serials

        expected = json.dumps(removed_nonremoved_serials[2])
        result = self.obj.RemoveEntitlementsByPoolIds.__wrapped__(self.obj, ["123"], {}, self.LOCALE)
        self.assertEqual(expected, result)

    def test_RemoveAllEntitlements(self):
        remove_all_entitlements_patch = mock.patch(
            "rhsmlib.services.entitlement.EntitlementService.remove_all_entitlements",
            name="remove_all_entitlements",
        )
        self.patches["remove_all_entitlements"] = remove_all_entitlements_patch.start()
        self.addCleanup(remove_all_entitlements_patch.stop)

        records = {"deletedRecords": 1}
        self.patches["remove_all_entitlements"].return_value = records

        expected = json.dumps(records)
        result = self.obj.RemoveAllEntitlements.__wrapped__(self.obj, {}, self.LOCALE)
        self.assertEqual(expected, result)



import os
import sys

from ..test_managercli import TestCliProxyCommand
from subscription_manager import managercli
from subscription_manager.entcertlib import CONTENT_ACCESS_CERT_TYPE
from subscription_manager.injection import provide, CERT_SORTER

from ..stubs import StubProductCertificate, StubEntitlementCertificate, StubProduct, StubCertSorter, StubPool
from ..fixture import Capture

from unittest.mock import patch, Mock, MagicMock


class TestListCommand(TestCliProxyCommand):
    command_class = managercli.ListCommand
    valid_date = "2018-05-01"

    def setUp(self):
        super(TestListCommand, self).setUp(False)
        self.indent = 1
        self.max_length = 40
        self.cert_with_service_level = StubEntitlementCertificate(
            StubProduct("test-product"), service_level="Premium"
        )
        self.cert_with_content_access = StubEntitlementCertificate(
            StubProduct("test-product"), entitlement_type=CONTENT_ACCESS_CERT_TYPE
        )
        argv_patcher = patch.object(sys, "argv", ["subscription-manager", "list"])
        argv_patcher.start()
        self.addCleanup(argv_patcher.stop)

    def _test_afterdate_option(self, argv, method, should_exit=True, expected_exit_code=0):
        msg = ""
        with patch.object(sys, "argv", argv):
            try:
                method()
            except SystemExit as e:
                self.assertEqual(
                    e.code,
                    expected_exit_code,
                    """Cli should have exited with code '{}', got '{}'""".format(expected_exit_code, e.code),
                )
                fail = False
            except Exception as e:
                fail = True
                msg = "Expected SystemExit, got '''{}'''".format(e)
            else:
                fail = should_exit
                if fail:
                    msg = "Expected SystemExit, No Exception was raised"

            if fail:
                self.fail(msg)

    def test_afterdate_option_bad_date(self):
        argv = ["subscription-manager", "list", "--all", "--available", "--afterdate", "not_a_real_date"]
        self._test_afterdate_option(argv, self.cc.main, expected_exit_code=os.EX_DATAERR)

    def test_afterdate_option_no_date(self):
        argv = ["subscription-manager", "list", "--all", "--available", "--afterdate"]
        # Error code of 2 is expected from optparse in this case.
        self._test_afterdate_option(argv, self.cc.main, expected_exit_code=2)

    def test_afterdate_option_missing_options(self):
        # Just missing "available"
        argv = ["subscription-manager", "list", "--afterdate", self.valid_date, "--all"]
        self._test_afterdate_option(argv, self.cc.main, expected_exit_code=os.EX_USAGE)

        # Missing both
        argv = ["subscription-manager", "list", "--afterdate", self.valid_date]
        self._test_afterdate_option(argv, self.cc.main, expected_exit_code=os.EX_USAGE)

    def test_afterdate_option_with_ondate(self):
        argv = ["subscription-manager", "list", "--afterdate", self.valid_date, "--ondate", self.valid_date]
        self._test_afterdate_option(argv, self.cc.main, expected_exit_code=os.EX_USAGE)

    @patch("subscription_manager.managerlib.get_available_entitlements")
    def test_afterdate_option_valid(self, es):
        def create_pool_list(*args, **kwargs):
            return [
                {
                    "productName": "dummy-name",
                    "productId": "dummy-id",
                    "providedProducts": [],
                    "id": "888888888888",
                    "management_enabled": True,
                    "attributes": [{"name": "is_virt_only", "value": "false"}],
                    "pool_type": "Some Type",
                    "quantity": "4",
                    "service_type": "",
                    "roles": "awsome server",
                    "service_level": "",
                    "usage": "Testing",
                    "addons": "ADDON1",
                    "contractNumber": "5",
                    "multi-entitlement": "false",
                    "startDate": "",
                    "endDate": "",
                    "suggested": "2",
                }
            ]

        es.return_value = create_pool_list()

        argv = ["subscription-manager", "list", "--all", "--available", "--afterdate", self.valid_date]
        self._test_afterdate_option(argv, self.cc.main, should_exit=False)

    @patch("subscription_manager.managerlib.get_available_entitlements")
    def test_none_wrap_available_pool_id(self, mget_ents):
        list_command = managercli.ListCommand()

        def create_pool_list(*args, **kwargs):
            return [
                {
                    "productName": "dummy-name",
                    "productId": "dummy-id",
                    "providedProducts": [],
                    "id": "888888888888",
                    "management_enabled": True,
                    "attributes": [{"name": "is_virt_only", "value": "false"}],
                    "pool_type": "Some Type",
                    "quantity": "4",
                    "service_type": "",
                    "roles": "awesome server",
                    "service_level": "",
                    "usage": "Production",
                    "addons": "",
                    "contractNumber": "5",
                    "multi-entitlement": "false",
                    "startDate": "",
                    "endDate": "",
                    "suggested": "2",
                }
            ]

        mget_ents.return_value = create_pool_list()

        with Capture() as cap:
            list_command.main(["--available"])
        self.assertTrue("888888888888" in cap.out)

    @patch("subscription_manager.managerlib.get_available_entitlements")
    def test_available_syspurpose_attr(self, mget_ents):
        list_command = managercli.ListCommand()

        def create_pool_list(*args, **kwargs):
            return [
                {
                    "productName": "dummy-name",
                    "productId": "dummy-id",
                    "providedProducts": [],
                    "id": "888888888888",
                    "management_enabled": True,
                    "attributes": [{"name": "is_virt_only", "value": "false"}],
                    "pool_type": "Some Type",
                    "quantity": "4",
                    "service_type": "",
                    "roles": "Awesome Server, Cool Server",
                    "service_level": "Premium",
                    "usage": "Production",
                    "addons": "ADDON1,ADDON2",
                    "contractNumber": "5",
                    "multi-entitlement": "false",
                    "startDate": "",
                    "endDate": "",
                    "suggested": "2",
                }
            ]

        mget_ents.return_value = create_pool_list()

        with Capture() as cap:
            list_command.main(["--available"])
        self.assertTrue("ADDON1\n" in cap.out)
        self.assertTrue("Awesome Server\n" in cap.out)
        self.assertTrue("Production" in cap.out)
        self.assertTrue("Premium" in cap.out)

    def test_print_consumed_no_ents(self):
        with Capture() as captured:
            self.cc.print_consumed()

        lines = captured.out.split("\n")
        self.assertEqual(len(lines) - 1, 1, "Error output consists of more than one line.")

    def test_list_installed(self):
        """
        Test output of 'subscription-manager list --installed'
        """
        installed_product_certs = [
            StubProductCertificate(product=StubProduct(name="test product", product_id="8675309")),
            StubProductCertificate(product=StubProduct(name="another test product", product_id="123456")),
        ]

        stub_sorter = StubCertSorter()

        for product_cert in installed_product_certs:
            product = product_cert.products[0]
            stub_sorter.installed_products[product.id] = product_cert

        provide(CERT_SORTER, stub_sorter)

        with Capture() as captured:
            list_command = managercli.ListCommand()
            list_command.main(["--installed"])
            assert "Product Name:" in captured.out
            assert "Product ID:" in captured.out
            assert "Version:" in captured.out
            assert "Arch:" in captured.out
            assert "Status:" in captured.out
            assert "Status Details:" in captured.out
            assert "Starts:" in captured.out
            assert "Ends:" in captured.out

    @patch("subscription_manager.cli_command.list.is_simple_content_access")
    def test_list_installed_sca_mode(self, is_simple_content_access_mock):
        """
        Test output of 'subscription-manager list --installed', when SCA mode is used
        """
        is_simple_content_access_mock.return_value = True

        installed_product_certs = [
            StubProductCertificate(product=StubProduct(name="test product", product_id="8675309")),
            StubProductCertificate(product=StubProduct(name="another test product", product_id="123456")),
        ]

        stub_sorter = StubCertSorter()

        for product_cert in installed_product_certs:
            product = product_cert.products[0]
            stub_sorter.installed_products[product.id] = product_cert

        provide(CERT_SORTER, stub_sorter)

        with Capture() as captured:
            list_command = managercli.ListCommand()
            list_command.main(["--installed"])
            assert "Product Name:" in captured.out
            assert "Product ID:" in captured.out
            assert "Version:" in captured.out
            assert "Arch:" in captured.out
            # Following attributes should not be printed in SCA mode, because it does not make
            # any sense to print them in SCA mode
            assert "Status:" not in captured.out
            assert "Status Details:" not in captured.out
            assert "Starts:" not in captured.out
            assert "Ends:" not in captured.out

    def test_list_installed_with_ctfilter(self):
        installed_product_certs = [
            StubProductCertificate(product=StubProduct(name="test product*", product_id="8675309")),
            StubProductCertificate(product=StubProduct(name="another(?) test\\product", product_id="123456")),
        ]

        test_data = [
            ("", (True, True)),
            ("input string", (False, False)),
            ("*product", (False, True)),
            ("*product*", (True, True)),
            ("*test pro*uct*", (True, False)),
            ("*test pro?uct*", (True, False)),
            ("*test pr*ct*", (True, False)),
            ("*test pr?ct*", (False, False)),
            ("*another*", (False, True)),
            ("*product\\*", (True, False)),
            ("*product?", (True, False)),
            ("*product?*", (True, False)),
            ("*(\\?)*", (False, True)),
            ("*test\\\\product", (False, True)),
        ]

        stub_sorter = StubCertSorter()

        for product_cert in installed_product_certs:
            product = product_cert.products[0]
            stub_sorter.installed_products[product.id] = product_cert

        provide(CERT_SORTER, stub_sorter)

        for (test_num, data) in enumerate(test_data):
            with Capture() as captured:
                list_command = managercli.ListCommand()
                list_command.main(["--installed", "--matches", data[0]])

            for (index, expected) in enumerate(data[1]):
                if expected:
                    self.assertTrue(
                        installed_product_certs[index].name in captured.out,
                        "Expected product was not found in output for test data %i" % test_num,
                    )
                else:
                    self.assertFalse(
                        installed_product_certs[index].name in captured.out,
                        "Unexpected product was found in output for test data %i" % test_num,
                    )

    def test_list_consumed_with_ctfilter(self):
        consumed = [
            StubEntitlementCertificate(
                product=StubProduct(name="Test Entitlement 1", product_id="123"),
                provided_products=["test product a", "beta product 1", "shared product", "troll* product?"],
            ),
            StubEntitlementCertificate(
                product=StubProduct(name="Test Entitlement 2", product_id="456"),
                provided_products=["test product b", "beta product 1", "shared product", "back\\slash"],
            ),
        ]

        test_data = [
            ("", (False, False)),
            ("test entitlement ?", (True, True)),
            ("*entitlement 1", (True, False)),
            ("*entitlement 2", (False, True)),
            ("input string", (False, False)),
            ("*product", (True, True)),
            ("*product*", (True, True)),
            ("shared pro*nopenopenope", (False, False)),
            ("*another*", (False, False)),
            ("*product\\?", (True, False)),
            ("*product ?", (True, True)),
            ("*product?*", (True, True)),
            ("*\\?*", (True, False)),
            ("*\\\\*", (False, True)),
            ("*k\\s*", (False, True)),
            ("*23", (True, False)),
            ("45?", (False, True)),
        ]

        for stubby in consumed:
            self.ent_dir.certs.append(stubby)

        for (test_num, data) in enumerate(test_data):
            with Capture() as captured:
                list_command = managercli.ListCommand()
                list_command.main(["--consumed", "--matches", data[0]])

            for (index, expected) in enumerate(data[1]):
                if expected:
                    self.assertTrue(
                        consumed[index].order.name in captured.out,
                        "Expected product was not found in output for test data %i" % test_num,
                    )
                else:
                    self.assertFalse(
                        consumed[index].order.name in captured.out,
                        "Unexpected product was found in output for test data %i" % test_num,
                    )

    def test_print_consumed_one_ent_one_product(self):
        product = StubProduct("product1")
        self.ent_dir.certs.append(StubEntitlementCertificate(product))
        self.cc.sorter = Mock()
        self.cc.sorter.get_subscription_reasons_map = Mock()
        self.cc.sorter.get_subscription_reasons_map.return_value = {}
        self.cc.print_consumed()

    def test_print_consumed_one_ent_no_product(self):
        self.ent_dir.certs.append(StubEntitlementCertificate(product=None))
        self.cc.sorter = Mock()
        self.cc.sorter.get_subscription_reasons_map = Mock()
        self.cc.sorter.get_subscription_reasons_map.return_value = {}
        self.cc.print_consumed()

    def test_print_consumed_prints_nothing_with_no_service_level_match(self):
        self.ent_dir.certs.append(self.cert_with_service_level)

        with Capture() as captured:
            self.cc.print_consumed(service_level="NotFound")

        lines = captured.out.split("\n")
        self.assertEqual(len(lines) - 1, 1, "Error output consists of more than one line.")

    def test_print_consumed_prints_enitlement_with_service_level_match(self):
        self.ent_dir.certs.append(self.cert_with_service_level)
        self.cc.sorter = Mock()
        self.cc.sorter.get_subscription_reasons_map = Mock()
        self.cc.sorter.get_subscription_reasons_map.return_value = {}
        self.cc.print_consumed(service_level="Premium")

    def test_print_consumed_ignores_content_access_cert(self):
        self.ent_dir.certs.append(self.cert_with_content_access)
        with Capture() as captured:
            self.cc.print_consumed(service_level="NotFound")

        lines = captured.out.split("\n")
        self.assertEqual(len(lines) - 1, 1, "Error output consists of more than one line.")

    def test_list_installed_with_pidonly(self):
        installed_product_certs = [
            StubProductCertificate(product=StubProduct(name="test product*", product_id="8675309")),
            StubProductCertificate(product=StubProduct(name="another(?) test\\product", product_id="123456")),
        ]

        stub_sorter = StubCertSorter()

        for product_cert in installed_product_certs:
            product = product_cert.products[0]
            stub_sorter.installed_products[product.id] = product_cert

        provide(CERT_SORTER, stub_sorter)

        try:
            with Capture() as captured:
                list_command = managercli.ListCommand()
                list_command.main(["--installed", "--pool-only"])

            self.fail("Expected error did not occur")
        except SystemExit:
            for cert in installed_product_certs:
                self.assertFalse(cert.products[0].id in captured.out)

    def test_list_consumed_with_pidonly(self):
        consumed = [
            StubEntitlementCertificate(
                product=StubProduct(name="Test Entitlement 1", product_id="123"),
                pool=StubPool("abc"),
                provided_products=["test product a", "beta product 1", "shared product", "troll* product?"],
            ),
            StubEntitlementCertificate(
                product=StubProduct(name="Test Entitlement 2", product_id="456"),
                pool=StubPool("def"),
                provided_products=["test product b", "beta product 1", "shared product", "back\\slash"],
            ),
        ]

        for stubby in consumed:
            self.ent_dir.certs.append(stubby)

        with Capture() as captured:
            list_command = managercli.ListCommand()
            list_command.main(["--consumed", "--pool-only"])

        for cert in consumed:
            self.assertFalse(cert.order.name in captured.out)
            self.assertTrue(cert.pool.id in captured.out)

    def test_list_consumed_syspurpose_attr_version34(self):
        """
        When version of entitlement certificate is 3.4, then subscription-manager should print syspurpose
        attributes from the certificate.
        """
        product = StubProduct("product1")
        ent_cert = StubEntitlementCertificate(product)
        ent_cert.order.usage = "Development"
        ent_cert.order.roles = ["SP Server", "SP Starter"]
        ent_cert.order.addons = ["ADDON1", "ADDON2"]
        ent_cert.version = MagicMock()
        ent_cert.version.major = 3
        ent_cert.version.minor = 4
        self.ent_dir.certs.append(ent_cert)
        self.cc.sorter = Mock()
        self.cc.sorter.get_subscription_reasons_map = Mock()
        self.cc.sorter.get_subscription_reasons_map.return_value = {}
        with Capture() as captured:
            self.cc.print_consumed()
            self.assertTrue("Add-ons:" in captured.out)
            self.assertTrue("ADDON1" in captured.out)
            self.assertTrue("ADDON2" in captured.out)
            self.assertTrue("Usage:" in captured.out)
            self.assertTrue("Development" in captured.out)
            self.assertTrue("Roles:" in captured.out)
            self.assertTrue("SP Server" in captured.out)
            self.assertTrue("SP Starter" in captured.out)

    def test_list_consumed_no_syspurpose_attr_version33(self):
        """
        When the version of certificate is older then 3.4, then do not print syspurpose attributes, because
        there cannot be any.
        """
        product = StubProduct("product1")
        ent_cert = StubEntitlementCertificate(product)
        ent_cert.version = MagicMock()
        ent_cert.version.major = 3
        ent_cert.version.minor = 3
        self.ent_dir.certs.append(ent_cert)
        self.cc.sorter = Mock()
        self.cc.sorter.get_subscription_reasons_map = Mock()
        self.cc.sorter.get_subscription_reasons_map.return_value = {}
        with Capture() as captured:
            self.cc.print_consumed()
            self.assertFalse("Add-ons:" in captured.out)
            self.assertFalse("Usage:" in captured.out)
            self.assertFalse("Roles:" in captured.out)



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import os

import rhsm.connection as connection

from subscription_manager.cli import system_exit
from subscription_manager.cli_command.cli import CliCommand
from subscription_manager.i18n import ugettext as _
from subscription_manager.overrides import Overrides, Override
from subscription_manager.printing_utils import columnize, echo_columnize_callback
from subscription_manager.repofile import manage_repos_enabled
from subscription_manager.utils import get_supported_resources


class OverrideCommand(CliCommand):
    def __init__(self):
        shortdesc = _("Manage custom content repository settings")
        super(OverrideCommand, self).__init__("repo-override", shortdesc, False)
        self.parser.add_argument(
            "--repo",
            dest="repos",
            action="append",
            metavar="REPOID",
            help=_("repository to modify (can be specified more than once)"),
        )
        self.parser.add_argument(
            "--remove",
            dest="removals",
            action="append",
            metavar="NAME",
            help=_(
                "name of the override to remove (can be specified more than once); used with --repo option."
            ),
        )
        self.parser.add_argument(
            "--add",
            dest="additions",
            action="append",
            metavar="NAME:VALUE",
            help=_(
                "name and value of the option to override separated by a colon "
                "(can be specified more than once); used with --repo option."
            ),
        )
        self.parser.add_argument(
            "--remove-all",
            action="store_true",
            help=_("remove all overrides; can be specific to a repository by providing --repo"),
        )
        self.parser.add_argument(
            "--list",
            action="store_true",
            help=_("list all overrides; can be specific to a repository by providing --repo"),
        )

    def _additions_colon_split(self):
        additions = {}
        for value in self.options.additions or {}:
            if value.strip() == "":
                system_exit(
                    os.EX_USAGE, _('You must specify an override in the form of "name:value" with --add.')
                )

            k, _colon, v = value.partition(":")
            if not v or not k:
                system_exit(os.EX_USAGE, _('--add arguments should be in the form of "name:value"'))

            additions[k] = v
        self.options.additions = additions

    def _validate_options(self):
        if self.options.additions:
            self._additions_colon_split()
        if self.options.additions or self.options.removals:
            if not self.options.repos:
                system_exit(os.EX_USAGE, _("Error: You must specify a repository to modify"))
            if self.options.remove_all or self.options.list:
                system_exit(
                    os.EX_USAGE, _("Error: You may not use --add or --remove with --remove-all and --list")
                )
        if self.options.list and self.options.remove_all:
            system_exit(os.EX_USAGE, _("Error: You may not use --list with --remove-all"))
        if self.options.repos and not (
            self.options.list or self.options.additions or self.options.removals or self.options.remove_all
        ):
            system_exit(
                os.EX_USAGE, _("Error: The --repo option must be used with --list or --add or --remove.")
            )
        if self.options.removals:
            stripped_removals = [removal.strip() for removal in self.options.removals]
            if "" in stripped_removals:
                system_exit(os.EX_USAGE, _("Error: You must specify an override name with --remove."))
        # If no relevant options were given, just show a list
        if not (
            self.options.repos
            or self.options.additions
            or self.options.removals
            or self.options.remove_all
            or self.options.list
        ):
            self.options.list = True

    def _do_command(self):
        self._validate_options()
        # Abort if not registered
        self.assert_should_be_registered()

        supported_resources = get_supported_resources()
        if "content_overrides" not in supported_resources:
            system_exit(
                os.EX_UNAVAILABLE, _("Error: The 'repo-override' command is not supported by the server.")
            )

        # update entitlement certificates if necessary. If we do have new entitlements
        # CertLib.update() will call RepoActionInvoker.update().
        self.entcertlib.update()
        # make sure the EntitlementDirectory singleton is refreshed
        self._request_validity_check()

        overrides = Overrides()

        if not manage_repos_enabled():
            print(_("Repositories disabled by configuration."))

        if self.options.list:
            results = overrides.get_overrides(self.identity.uuid)
            if results:
                self._list(results, self.options.repos)
            else:
                print(_("This system does not have any content overrides applied to it."))
            return

        if self.options.additions:
            repo_ids = [repo.id for repo in overrides.repo_lib.get_repos(apply_overrides=False)]
            to_add = [
                Override(repo, name, value)
                for repo in self.options.repos
                for name, value in list(self.options.additions.items())
            ]
            try:
                results = overrides.add_overrides(self.identity.uuid, to_add)
            except connection.RestlibException as ex:
                if ex.code == 400:
                    # blocklisted overrides specified.
                    # Print message and return a less severe code.
                    system_exit(1, ex)
                else:
                    raise ex

            # Print out warning messages if the specified repo does not exist in the repo file.
            for repo in self.options.repos:
                if repo not in repo_ids:
                    print(
                        _(
                            "Repository '{repo}' does not currently exist, but the override has been added."
                        ).format(repo=repo)
                    )

        if self.options.removals:
            to_remove = [
                Override(repo, item) for repo in self.options.repos for item in self.options.removals
            ]
            results = overrides.remove_overrides(self.identity.uuid, to_remove)
        if self.options.remove_all:
            results = overrides.remove_all_overrides(self.identity.uuid, self.options.repos)

        # Update the cache and refresh the repo file.
        overrides.update(results)

    def _list(self, all_overrides, specific_repos):
        overrides = {}
        for override in all_overrides:
            repo = override.repo_id
            name = override.name
            value = override.value
            # overrides is a hash of hashes.  Like this: {'repo_x': {'enabled': '1', 'gpgcheck': '1'}}
            overrides.setdefault(repo, {})[name] = value

        to_show = set(overrides.keys())
        if specific_repos:
            specific_repos = set(specific_repos)
            for r in specific_repos.difference(to_show):
                print(_("Nothing is known about '{r}'").format(r=r))
            # Take the intersection of the sets
            to_show &= specific_repos

        for repo in sorted(to_show):
            print(_("Repository: {repo}").format(repo=repo))
            repo_data = sorted(list(overrides[repo].items()), key=lambda x: x[0])
            # Split the list of 2-tuples into a list of names and a list of keys
            names, values = list(zip(*repo_data))
            names = ["{x}:".format(x=x) for x in names]
            print(columnize(names, echo_columnize_callback, *values, indent=2) + "\n")



# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

"""
Setup for cloud-what
"""

from setuptools import setup

setup(
    name="cloud-what",
    version="1.29.14",
    description="Detection of public cloud provider and gathering metadata from IMDS servers",
    author="Jiri Hnidek",
    author_email="chainsaw@redhat.com",
    url="https://github.com/candlepin/subscription-manager/",
    license="GPLv2",
    long_description="""cloud-what enables to detect cloud provider using information provided
    SM BIOS. The package tries to use dmidecode and virt-what for this purpose. This package
    also allows to gather metadata and signature from IMDS servers provided by public cloud
    providers. Three cloud providers are supported at this moment: AWS, Azure and GCP.""",
    packages=["cloud-what"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GPL License",
        "License :: OSI Approved :: Python Software Foundation License",
        "Operating System :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=["requests"],
)



# This module has been originally modified and enhanced from Red Hat Update
# Agent's config module.
#
# Copyright (c) 2010 - 2012 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

import sys
import os
import logging
from iniparse import SafeConfigParser
from iniparse.compat import NoOptionError, InterpolationMissingOptionError, NoSectionError
import re
import tempfile
from typing import Dict, List, Optional, Tuple
from subscription_manager.i18n import ugettext as _

CONFIG_ENV_VAR = "RHSM_CONFIG"

DEFAULT_CONFIG_DIR = "/etc/rhsm/"
HOST_CONFIG_DIR = "/etc/rhsm-host/"  # symlink inside docker containers
DEFAULT_CONFIG_PATH = "%srhsm.conf" % DEFAULT_CONFIG_DIR
DEFAULT_PROXY_PORT = "3128"
DEFAULT_SERVER_TIMEOUT = "180"

# Defaults for connecting to RHSM, used to "reset" the configuration file
# if requested by the user:
DEFAULT_HOSTNAME = "subscription.rhsm.redhat.com"
DEFAULT_PORT = "443"
DEFAULT_PREFIX = "/subscription"

DEFAULT_CDN_HOSTNAME = "cdn.redhat.com"
DEFAULT_CDN_PORT = "443"
DEFAULT_CDN_PREFIX = "/"

DEFAULT_CA_CERT_DIR = "/etc/rhsm/ca/"

DEFAULT_ENT_CERT_DIR = "/etc/pki/entitlement"
HOST_ENT_CERT_DIR = "/etc/pki/entitlement-host"

SERVER_DEFAULTS = {
    "hostname": DEFAULT_HOSTNAME,
    "prefix": DEFAULT_PREFIX,
    "port": DEFAULT_PORT,
    "server_timeout": DEFAULT_SERVER_TIMEOUT,
    "insecure": "0",
    "proxy_hostname": "",
    "proxy_scheme": "http",
    "proxy_user": "",
    "proxy_port": "",
    "proxy_password": "",
    "no_proxy": "",
}
RHSM_DEFAULTS = {
    "baseurl": "https://" + DEFAULT_CDN_HOSTNAME,
    "repomd_gpg_url": "",
    "ca_cert_dir": DEFAULT_CA_CERT_DIR,
    "repo_ca_cert": "%(ca_cert_dir)sredhat-uep.pem",
    "productcertdir": "/etc/pki/product",
    "entitlementcertdir": DEFAULT_ENT_CERT_DIR,
    "consumercertdir": "/etc/pki/consumer",
    "manage_repos": "1",
    "full_refresh_on_yum": "0",
    "report_package_profile": "1",
    "plugindir": "/usr/share/rhsm-plugins",
    "pluginconfdir": "/etc/rhsm/pluginconf.d",
    "auto_enable_yum_plugins": "1",
    "package_profile_on_trans": "0",
    "inotify": "1",
    "progress_messages": "1",
}

RHSMCERTD_DEFAULTS = {
    "certcheckinterval": "240",
    "autoattachinterval": "1440",
    "splay": "1",
    "disable": "0",
    "auto_registration": "0",
    "auto_registration_interval": "60",
}

LOGGING_DEFAULTS = {
    "default_log_level": "INFO",
}

# Defaults are applied to each section in the config file.
DEFAULTS = {
    "server": SERVER_DEFAULTS,
    "rhsm": RHSM_DEFAULTS,
    "rhsmcertd": RHSMCERTD_DEFAULTS,
    "logging": LOGGING_DEFAULTS,
}


log = logging.getLogger(__name__)


def in_container() -> bool:
    """
    Are we running in a docker container or not?
    """
    # For development in containers we must be able to turn container detection
    # off
    if os.environ.get("SMDEV_CONTAINER_OFF", ""):
        return False

    def in_ocp() -> bool:
        """
        Is the system running as pod in OCP (OpenShift Container Platform)?

        Check some of the canonical environment variables set by Kubernets
        (on which OCP is based):
        https://kubernetes.io/docs/concepts/containers/container-environment/
        in particular, look for the "kubernetes" default service

        "container=oci" is generally set by Red Hat-based containers.
        """
        return (
            os.environ.get("KUBERNETES_PORT", "") != ""
            and os.environ.get("KUBERNETES_SERVICE_HOST", "") != ""
            and os.environ.get("KUBERNETES_SERVICE_PORT", "") != ""
            and os.environ.get("container", "") == "oci"
        )

    # Known locations to check for as an easy way to detect whether
    # we are running in a container; note that pods in OCP are not
    # considered containers but standalone systems
    locations: List[str] = [
        # podman:
        # https://github.com/containers/podman/issues/6192
        # https://github.com/containers/podman/issues/3586#issuecomment-661918679
        "/run/.containerenv",
        # docker:
        # https://github.com/moby/moby/issues/18355
        "/.dockerenv",
        # The host rhsm configuration was shared with us, so assume
        # we must be running in a container
        HOST_CONFIG_DIR,
    ]
    for fn in locations:
        if os.path.exists(fn):
            log.debug(f"in_container(): found '{fn}', may be a container")
            is_ocp = in_ocp()
            if is_ocp:
                log.debug("in_container(): found kubernetes/OCP environment, not considering container")
                return False
            return True
    return False


class RhsmConfigParser(SafeConfigParser):
    """Config file parser for rhsm configuration."""

    # defaults unused but kept to preserve compatibility
    def __init__(self, config_file: Optional[str] = None, defaults=None):
        self.config_file: str = config_file
        SafeConfigParser.__init__(self)
        self.read(self.config_file)

    def read(self, file_names: Optional[List[str]] = None) -> List[str]:
        """
        Read configuration files. When configuration files are not specified, then read self.config_file
        :param file_names: list of configuration files
        :return: number of configuration files read
        """
        if file_names is None:
            return super(RhsmConfigParser, self).read(self.config_file)
        else:
            return super(RhsmConfigParser, self).read(file_names)

    def save(self, config_file: Optional[str] = None) -> None:
        """Writes config file to storage."""
        # Use self.config_file if config_file is not provided in method arguments.
        config_file: str = config_file or self.config_file

        rhsm_conf_dir: str = os.path.dirname(config_file)

        # When /etc/rhsm does not exist, then try to create it
        if os.path.isdir(rhsm_conf_dir) is False:
            os.makedirs(rhsm_conf_dir)

        # Create a temporary file to write config data to it and
        # rename the file to the expected config file name after successfully
        # writing all config data.
        # Refer to BZ 1719725: https://bugzilla.redhat.com/show_bug.cgi?id=1719725
        with tempfile.NamedTemporaryFile(mode="w", dir=rhsm_conf_dir, delete=False) as fo:
            self.write(fo)
            fo.flush()
            mode: int
            try:
                mode = os.stat(config_file).st_mode
            except IOError:
                mode = 0o644
            os.rename(fo.name, config_file)
            os.chmod(config_file, mode)

    def get(self, section: str, prop: str) -> str:
        """Get a value from rhsm config.

        :param section: config file section
        :param prop: what config property to find, the config item name
        :return: The string value of the config item.

        If config item exists, but is not set, an empty string is returned.
        """
        try:
            return SafeConfigParser.get(self, section, prop)
        except InterpolationMissingOptionError:
            # if there is an interpolation error, resolve it
            raw_val: str = super(RhsmConfigParser, self).get(section, prop, True)
            interpolations: List[str] = re.findall(r"%\((.*?)\)s", raw_val)
            changed: bool = False
            for interp in interpolations:
                # Defaults aren't interpolated by default, so bake them in as necessary
                # has_option throws an exception if the section doesn't exist,
                # but at this point we know it does.
                if self.has_option(section, interp):
                    super(RhsmConfigParser, self).set(section, interp, self.get(section, interp))
                    changed = True
            if changed:
                # Now that we have the required values, we can interpolate
                return self.get(section, prop)
            # If nothing has been changed (we couldn't fix it) re-raise the exception
            raise
        except (NoOptionError, NoSectionError) as er:
            try:
                return DEFAULTS[section][prop.lower()]
            except KeyError:
                # re-raise the NoOptionError, not the key error
                raise er

    def set(self, section: str, name: str, value: str) -> None:
        try:
            # If the value doesn't exist, or isn't equal, write it
            if self.get(section, name) != value:
                raise NoOptionError
        except Exception:
            if not self.has_section(section):
                self.add_section(section)
            super(RhsmConfigParser, self).set(section, name, value)
        if section == "logging" and name == "default_log_level":
            self.is_log_level_valid(value)

    def is_log_level_valid(self, value: str, print_warning: bool = True) -> bool:
        """
        Check if provided default_log_level value is valid or not
        :param value: value of default_log_level
        :param print_warning: print warning, when provided value is not valid
        :return: True, when value is valid. Otherwise return False
        """
        valid: List[str] = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
        if value not in valid + ["NOTSET"]:
            if print_warning is True:
                print(
                    _("Invalid Log Level: {lvl}, setting to INFO for this run.".format(lvl=value)),
                    file=sys.stderr,
                )
                print(
                    _(
                        "Please use:  subscription-manager config --logging.default_log_level=<Log Level> to "
                        "set the default_log_level to a valid value."
                    ),
                    file=sys.stderr,
                )
                valid_str = ", ".join(valid)
                print(_("Valid Values: {valid_str}".format(valid_str=valid_str)), file=sys.stderr)
            return False
        return True

    def get_int(self, section: str, prop: str) -> Optional[int]:
        """Get an int value from the config.

        :param section: the config section
        :param prop: the config item name
        :return:
            An int cast from the string read from.
            If config item is unset, return None.
        :raises ValueError:
            If the config value found can not be coerced into an int
        """
        value_string: str = self.get(section, prop)
        if value_string == "":
            return None
        try:
            value_int = int(value_string)
            # we could also try to handle port name
            # strings (ie, 'http') here with getservbyname
        except (ValueError, TypeError):
            raise ValueError("Section: %s, Property: %s - Integer value expected" % (section, prop))
        return value_int

    # Overriding this method to address
    # http://code.google.com/p/iniparse/issues/detail?id=9
    def defaults(self) -> Dict[str, str]:
        result: List[Tuple[str, str]] = []
        for section in DEFAULTS:
            result += [(key, value) for (key, value) in list(DEFAULTS[section].items())]
        return dict(result)

    def sections(self) -> List[str]:
        result: List[str] = super(RhsmConfigParser, self).sections()
        for section in DEFAULTS:
            if section not in result:
                result.append(section)
        return result

    def has_option(self, section: str, prop: str) -> bool:
        try:
            self.get(section, prop)
            return True
        except NoOptionError:
            return False

    def items(self, section: str) -> List[Tuple[str, str]]:
        result: Dict[str, str] = {}
        for key in DEFAULTS.get(section, {}):
            result[key] = DEFAULTS[section][key]
        if self.has_section(section):
            super_result: List[str] = super(RhsmConfigParser, self).options(section)
            for key in super_result:
                if self.get(section, key) and len(self.get(section, key).strip()) > 0:
                    result[key] = self.get(section, key)
        return list(result.items())

    def options(self, section: str) -> List[str]:
        # This is necessary because with the way we handle defaults, parser.has_section('xyz')
        # will return True if 'xyz' exists only in the defaults but parser.options('xyz')
        #  will throw an exception.
        items = set()
        for key in DEFAULTS.get(section, {}):
            items.add(key)
        if self.has_section(section):
            super_result = super(RhsmConfigParser, self).options(section)
            items.update(super_result)
        return list(items)

    def is_default(self, section: str, prop: str, value: str) -> bool:
        if self.get_default(section, prop) == value:
            return True
        return False

    def has_default(self, section: str, prop: str) -> bool:
        return section in DEFAULTS and prop.lower() in DEFAULTS[section]

    def get_default(self, section: str, prop: str) -> Optional[str]:
        if self.has_default(section, prop.lower()):
            return DEFAULTS[section][prop.lower()]
        return None


class RhsmHostConfigParser(RhsmConfigParser):
    """
    Sub-class of config parser automatically loaded when we detect that
    we're running in a container environment.

    Host config is shared with containers as /etc/rhsm-host. However the
    rhsm.conf within will still be referencing /etc/rhsm for a couple
    properties. (ca_cert_dir, repo_ca_cert)

    Instead we load config file normally, and assume to replace occurrences
    of /etc/rhsm with /etc/rhsm-host in these properties.

    A similar adjustment is necessary for /etc/pki/entitlement-host if
    present.
    """

    def __init__(self, config_file: Optional[str] = None, defaults=None) -> None:
        super().__init__(config_file, defaults)

        # Override the ca_cert_dir and repo_ca_cert if necessary:
        ca_cert_dir: str = self.get("rhsm", "ca_cert_dir")
        repo_ca_cert: str = self.get("rhsm", "repo_ca_cert")

        ca_cert_dir = ca_cert_dir.replace(DEFAULT_CONFIG_DIR, HOST_CONFIG_DIR)
        repo_ca_cert = repo_ca_cert.replace(DEFAULT_CONFIG_DIR, HOST_CONFIG_DIR)
        self.set("rhsm", "ca_cert_dir", ca_cert_dir)
        self.set("rhsm", "repo_ca_cert", repo_ca_cert)

        # Similarly if /etc/pki/entitlement-host exists, override this too.
        # If for some reason the host config is pointing to another directory
        # we leave the config setting alone, our tooling isn't going to be
        # able to handle it anyhow.
        if os.path.exists(HOST_ENT_CERT_DIR):
            ent_cert_dir = self.get("rhsm", "entitlementcertdir")
            if ent_cert_dir == DEFAULT_ENT_CERT_DIR or ent_cert_dir == DEFAULT_ENT_CERT_DIR + "/":
                ent_cert_dir = HOST_ENT_CERT_DIR
            self.set("rhsm", "entitlementcertdir", ent_cert_dir)


CFG: Optional[RhsmConfigParser] = None


def get_config_parser() -> RhsmConfigParser:
    """
    Get an :class:`RhsmConfig` instance

    Will use the first config file defined in the following list:
    - /etc/rhsm-host/rhsm.conf if it exists (only in containers)
    - /etc/rhsm/rhsm.conf
    """
    global CFG

    try:
        CFG = CFG
    except NameError:
        CFG = None

    if CFG is None:
        # Load alternate config file implementation if we detect that we're
        # running in a container.
        if in_container():
            CFG = RhsmHostConfigParser(config_file=os.path.join(HOST_CONFIG_DIR, "rhsm.conf"))
        else:
            CFG = RhsmConfigParser(config_file=DEFAULT_CONFIG_PATH)

    return CFG


# Deprecated but still in use by other applications
def initConfig(configFile=None):
    return get_config_parser()



import unittest

from rhsmlib.facts import pkg_arches
from unittest.mock import patch


class TestSupportedArchesCollector(unittest.TestCase):
    @staticmethod
    def helper_dpkg_no_foreign(*args, **kwargs):
        if args[0][1] == "--print-architecture":
            return "amd64\n".encode("UTF-8")
        return "".encode("UTF-8")

    @staticmethod
    def helper_dpkg_foreign(*args, **kwargs):
        if args[0][1] == "--print-architecture":
            return "amd64\n".encode("UTF-8")
        elif args[0][1] == "--print-foreign-architectures":
            return "i386\n".encode("UTF-8")
        return "".encode("UTF-8")

    @patch("subprocess.check_output")
    def test_single_arch_on_debian(self, MockCheckOutput):
        collector = pkg_arches.SupportedArchesCollector(collected_hw_info={"distribution.name": "Debian"})
        MockCheckOutput.side_effect = self.helper_dpkg_no_foreign
        fact = collector.get_all()
        self.assertEqual(fact["supported_architectures"], "amd64")

    @patch("subprocess.check_output")
    def test_single_arch_on_ubuntu(self, MockCheckOutput):
        collector = pkg_arches.SupportedArchesCollector(collected_hw_info={"distribution.name": "Ubuntu"})
        MockCheckOutput.side_effect = self.helper_dpkg_no_foreign
        fact = collector.get_all()
        self.assertEqual(fact["supported_architectures"], "amd64")

    @patch("subprocess.check_output")
    def test_multi_arch_on_ubuntu(self, MockCheckOutput):
        collector = pkg_arches.SupportedArchesCollector(collected_hw_info={"distribution.name": "Debian"})
        MockCheckOutput.side_effect = self.helper_dpkg_foreign
        fact = collector.get_all()
        self.assertEqual(fact["supported_architectures"], "amd64,i386")

    @patch("subprocess.check_output")
    def test_multi_arch_on_debian(self, MockCheckOutput):
        collector = pkg_arches.SupportedArchesCollector(collected_hw_info={"distribution.name": "Ubuntu"})
        MockCheckOutput.side_effect = self.helper_dpkg_foreign
        fact = collector.get_all()
        self.assertEqual(fact["supported_architectures"], "amd64,i386")

    @patch("subprocess.check_output")
    def test_none_arch_on_redhat(self, MockCheckOutput):
        collector = pkg_arches.SupportedArchesCollector(collected_hw_info={"distribution.name": "RedHat"})
        MockCheckOutput.side_effect = self.helper_dpkg_no_foreign
        fact = collector.get_all()
        self.assertTrue("supported_architectures" not in fact)



# Copyright (c) 2020 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

"""
This is module implementing detector and metadata collector of virtual machine
running on Google Cloud Platform.
"""

import logging
import time
import base64
import json

from typing import Union

from cloud_what._base_provider import BaseCloudProvider


log = logging.getLogger(__name__)


class GCPCloudProvider(BaseCloudProvider):
    """
    Class for GCP cloud provider

    Collector of Google Cloud Platform metadata. Verification of instance
    identity is described in this document:

    https://cloud.google.com/compute/docs/instances/verifying-instance-identity
    """

    CLOUD_PROVIDER_ID = "gcp"

    # The "audience" should be some unique URI agreed upon by both the instance
    # and the system verifying the instance's identity. For example, the
    # audience could be a URL for the connection between the two systems.
    # In fact this string could be anything.
    # TODO: use some more generic URL here and move setting this RHSM specific URL to subscription-manager
    AUDIENCE = "https://subscription.rhsm.redhat.com:443/subscription"

    # Google uses little bit different approach. It provides everything in JSON Web Token (JWT)
    CLOUD_PROVIDER_METADATA_URL = None

    CLOUD_PROVIDER_METADATA_URL_TEMPLATE = (
        "http://metadata/computeMetadata/v1/instance/service-accounts/default/identity"
        "?audience={audience}&format=full&licenses=TRUE"
    )

    # Token (metadata) expires usually within one hour. This is default value. Real value
    # is computed from expiration time contained in the JWT token.
    CLOUD_PROVIDER_METADATA_TTL = 3600

    # Threshold time in seconds
    THRESHOLD = 10

    CLOUD_PROVIDER_TOKEN_TTL = CLOUD_PROVIDER_METADATA_TTL

    CLOUD_PROVIDER_METADATA_TYPE = "text/html"

    CLOUD_PROVIDER_SIGNATURE_URL = None

    CLOUD_PROVIDER_SIGNATURE_TYPE = None

    HTTP_HEADERS = {
        "User-Agent": "cloud-what/1.0",
        "Metadata-Flavor": "Google",
    }

    # Metadata are provided in JWT token and this token is valid for one hour.
    # Thus it is save to cache this token (see CLOUD_PROVIDER_METADATA_TTL,
    # self._metadata_token_ctime and self._metadata_token)
    TOKEN_CACHE_FILE = "/var/cache/cloud-what/gcp_token.json"

    # Nothing to cache for this cloud provider
    SIGNATURE_CACHE_FILE = None

    def __init__(self, hw_info, audience_url=None):
        """
        Initialize instance of GCPCloudDetector
        """
        super(GCPCloudProvider, self).__init__(hw_info)

        # Metadata URL can have default or custom "audience"
        if audience_url is not None:
            self.CLOUD_PROVIDER_METADATA_URL = self.CLOUD_PROVIDER_METADATA_URL_TEMPLATE.format(
                audience=audience_url
            )
        else:
            self.CLOUD_PROVIDER_METADATA_URL = self.CLOUD_PROVIDER_METADATA_URL_TEMPLATE.format(
                audience=self.AUDIENCE
            )

    def is_running_on_cloud(self):
        """
        Try to guess if cloud provider is GCP using collected hardware information (output of dmidecode,
        virt-what, etc.)
        :return: True, when we detected sign of GCP in hardware information; Otherwise return False
        """

        # The system has to be VM
        if self.is_vm() is False:
            return False
        # This is valid for virtual machines running on Google Cloud Platform
        if "dmi.bios.vendor" in self.hw_info and "google" in self.hw_info["dmi.bios.vendor"].lower():
            return True
        # In other cases return False
        return False

    def is_likely_running_on_cloud(self):
        """
        Return non-zero value, when the machine is virtual machine and it is running on kvm and
        some google string can be found in output of dmidecode
        :return: Float value representing probability that vm is running on GPC
        """
        probability = 0.0

        # When the machine is not virtual machine, then there is probably zero chance that the machine
        # is running on GPC
        if self.is_vm() is False:
            return 0.0

        # We know that GCP uses only KVM at the end of 2020
        if "virt.host_type" in self.hw_info and "kvm" in self.hw_info["virt.host_type"]:
            probability += 0.3

        # Try to find "Google" or "gcp" keywords in output of dmidecode
        found_google = False
        found_gcp = False
        for hw_item in self.hw_info.values():
            if type(hw_item) != str:
                continue
            if "google" in hw_item.lower():
                found_google = True
            elif "gcp" in hw_item.lower():
                found_gcp = True
        if found_google is True:
            probability += 0.3
        if found_gcp is True:
            probability += 0.1

        return probability

    def _get_metadata_from_cache(self) -> Union[str, None]:
        """
        Try to get metadata (JWT token) from the cache file
        :return: String with cached token or None
        """
        return super(GCPCloudProvider, self)._get_token_from_cache_file()

    def _get_data_from_server(self, data_type: str, url: str, headers: dict = None) -> Union[str, None]:
        """
        Try to get data from metadata server
        """
        return super(GCPCloudProvider, self)._get_data_from_server(data_type, url, headers)

    def _get_ttl_from_metadata(self, metadata: dict, ctime: float) -> Union[None, float]:
        """
        Try to compute TTL from information provided in token metadata
        :param metadata: Dictionary with metadata
        :param ctime: Current time
        :return: TTL of token (time is in seconds from begin of epoch)
        """

        ttl = self.CLOUD_PROVIDER_TOKEN_TTL
        exp_time = None
        iat_time = None

        # Try to get expiration time from token
        if metadata is not None and "exp" in metadata:
            exp_time = metadata["exp"]
            # Decrease the ttl of value of threshold to expire token little sooner to be
            # sure that we do not use outdated token from cache
            ttl = exp_time - ctime - self.THRESHOLD
        else:
            log.debug(
                "GCP JWT token does not contain exp time, "
                f"using default TTL: {self.CLOUD_PROVIDER_TOKEN_TTL}"
            )

        # Try to get "ait" time (time, when the token was created)
        if metadata is not None and "iat" in metadata:
            # We try to test if there is not too big difference between creation time of JWT
            # token and current time (more than 10 seconds), because it would be strange,
            # and it could cause some other issues. Such situation could be caused e.g. by
            # wrong or missing NTP server, too big delay on connection, etc.
            iat_time = metadata["iat"]
            if abs(ctime - iat_time) > self.THRESHOLD:
                log.warning(f"Too big diff between ctime: {ctime} and GCP JWT token iat: {iat_time}")
        else:
            log.debug("GCP JWT token does not contain iat time")

        log.debug(f"GCP JWT token, exp: {exp_time}, iat: {iat_time}, TTL: {ttl}")

        return ttl

    def _get_metadata_from_server(self, headers: dict = None) -> Union[str, None]:
        """
        GCP metadata server returns only one file called token
        :return: String with token or None
        """
        token = self._get_data_from_server(
            data_type="token", url=self.CLOUD_PROVIDER_METADATA_URL, headers=headers
        )

        if token is not None:
            ttl = self.CLOUD_PROVIDER_TOKEN_TTL
            ctime: float = time.time()

            # Try to get expiration time from JWT token and the time, when the token was created
            jose_header, metadata, encoded_signature = self.decode_jwt(token)
            if metadata is not None:
                try:
                    metadata = json.loads(metadata)
                except json.JSONDecodeError as err:
                    log.error(f"Unable to decode metadata from GCP JWT token: {err}")
                else:
                    ttl = self._get_ttl_from_metadata(metadata, ctime)

            self._token = token
            self._token_ctime = ctime
            self._token_ttl = ttl
            self._write_token_to_cache_file()

        return token

    def _get_signature_from_server(self) -> None:
        """
        Google returns everything in one file.
        """
        return None

    def _get_signature_from_cache_file(self) -> None:
        """
        Really no need to cache signature
        """
        return None

    def get_signature(self) -> str:
        """
        Google returns everything in one file. No need to try to get signature.
        :return Empty string
        """
        return ""

    def get_metadata(self) -> Union[str, None]:
        """
        Try to get metadata from in-memory cache, cache or cloud provider server
        :return: String with metadata or None
        """
        if self._is_in_memory_cached_token_valid() is True:
            return self._token
        return super(GCPCloudProvider, self).get_metadata()

    def set_audience(self, audience: str) -> None:
        """
        Set audience unique identifier (usually some URL). This ID is used in HTTP request
        to GCP IMDS server
        :param audience: unique identifier
        :return: None
        """
        self.AUDIENCE = audience

    @staticmethod
    def decode_jwt(jwt_token: str) -> tuple:
        """
        Try to decode metadata stored in JWT token described in this RFC:
        https://tools.ietf.org/html/rfc7519
        :param jwt_token: string representing JWT token
        :return:
            tuple with: string representing header, string representing
            metadata and base64 encoded signature
        """
        # Get the actual payload part: [0] is JOSE header, [1] is metadata and [2] is signature
        parts = jwt_token.split(".")
        if len(parts) >= 3:
            encoded_jose_header = parts[0]
            encoded_metadata = parts[1]
            encoded_signature = parts[2]
            # Add some extra padding, JWT tokens have padding trimmed
            # See https://stackoverflow.com/a/49459036
            encoded_jose_header += "==="
            encoded_metadata += "==="
            encoded_signature += "==="
            # Decode only header and metadata, not signature
            try:
                jose_header = base64.b64decode(encoded_jose_header).decode("utf-8")
            except UnicodeDecodeError as err:
                log.error(f"Unable to decode JWT JOSE header: {err}")
                jose_header = None
            try:
                metadata = base64.b64decode(encoded_metadata).decode("utf-8")
            except UnicodeDecodeError as err:
                log.error(f"Unable to decode JWT metadata: {err}")
                metadata = None
            return jose_header, metadata, encoded_signature
        else:
            log.warning("JWT token with wrong format")
            return None, None, None


# Note about GCP token
# --------------------
#
# It is possible to verify token, but is not easy to do it on RHEL, because it requires
# special Python packages that are not available on RHEL. It is recommended to create
# virtual environment:
#
# $ python3 -m venv env
#
# Activate virtual environment:
#
# $ source env/bin/activate
#
# Install required packages:
#
# $ pip install --upgrade google-auth
# $ pip install requests
#
# Run following Python script:
#
# ```python
# from cloud_what.providers.gcp import GCPCloudCollector
# # Import libraries for token verification
# import google.auth.transport.requests
# from google.oauth2 import id_token
# # Get token
# token = GCPCloudCollector().get_metadata()
# # Verify token signature and store the token payload
# request = google.auth.transport.requests.Request()
# payload = id_token.verify_token(token, request=request, audience=GCPCloudCollector.AUDIENCE)
# print(payload)
# ```


def _smoke_test():
    """
    Simple smoke tests of GCP detector and collector
    """
    # Gather only information about hardware and virtualization
    from rhsmlib.facts.host_collector import HostCollector

    import sys

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    root.addHandler(handler)

    facts = {}
    facts.update(HostCollector().get_all())
    gcp_cloud_provider = GCPCloudProvider(facts)
    result = gcp_cloud_provider.is_running_on_cloud()
    probability = gcp_cloud_provider.is_likely_running_on_cloud()
    print(">>> debug <<< result: %s, %6.3f" % (result, probability))
    if result is True:
        # 1. using default audience
        token = gcp_cloud_provider.get_metadata()
        print(f">>> debug <<< 1. token: {token}")
        # 2. using some custom audience
        gcp_cloud_provider = GCPCloudProvider(facts, audience_url="https://localhost:8443/candlepin")
        token = gcp_cloud_provider.get_metadata()
        print(f">>> debug <<< 2. token: {token}")


# Some temporary smoke testing code. You can test this module using:
# sudo PYTHONPATH=./src python3 -m cloud_what.providers.gcp
if __name__ == "__main__":
    _smoke_test()



import threading
import time
import unittest

from unittest.mock import patch
from rhsm.utils import (
    remove_scheme,
    get_env_proxy_info,
    ServerUrlParseErrorEmpty,
    ServerUrlParseErrorNone,
    ServerUrlParseErrorPort,
    ServerUrlParseErrorScheme,
    ServerUrlParseErrorJustScheme,
    has_bad_scheme,
    has_good_scheme,
    parse_url,
    cmd_name,
    singleton,
    call_once,
    lock,
)
from rhsm.config import DEFAULT_PORT, DEFAULT_PREFIX, DEFAULT_HOSTNAME


class TestParseServerInfo(unittest.TestCase):
    def test_fully_specified(self):
        local_url = "myhost.example.com:900/myapp"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual("900", port)
        self.assertEqual("/myapp", prefix)

    def test_hostname_only(self):
        local_url = "myhost.example.com"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual(None, port)
        self.assertEqual(None, prefix)

    def test_hostname_port(self):
        local_url = "myhost.example.com:500"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual("500", port)
        self.assertEqual(None, prefix)

    def test_hostname_prefix(self):
        local_url = "myhost.example.com/myapp"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual(None, port)
        self.assertEqual("/myapp", prefix)

    def test_hostname_slash_no_prefix(self):
        local_url = "http://myhost.example.com/"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual(None, port)
        self.assertEqual("/", prefix)

    def test_hostname_just_slash(self):
        local_url = "/"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual(None, hostname)
        self.assertEqual(None, port)
        self.assertEqual("/", prefix)

    def test_hostname_just_slash_with_defaults(self):
        local_url = "/"
        (username, password, hostname, port, prefix) = parse_url(
            local_url, default_hostname=DEFAULT_HOSTNAME, default_port=DEFAULT_PORT
        )
        self.assertEqual(DEFAULT_HOSTNAME, hostname)
        self.assertEqual(DEFAULT_PORT, port)
        self.assertEqual("/", prefix)

    def test_hostname_nested_prefix(self):
        local_url = "myhost.example.com/myapp/subapp"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual(None, port)
        self.assertEqual("/myapp/subapp", prefix)

    def test_hostname_nothing(self):
        local_url = ""
        self.assertRaises(ServerUrlParseErrorEmpty, parse_url, local_url)

    def test_hostname_none(self):
        local_url = None
        self.assertRaises(ServerUrlParseErrorNone, parse_url, local_url)

    def test_hostname_with_scheme(self):
        # this is the default, so test it here
        local_url = "https://subscription.rhsm.redhat.com/subscription"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("subscription.rhsm.redhat.com", hostname)
        self.assertEqual(None, port)
        self.assertEqual("/subscription", prefix)

    def test_hostname_with_scheme_no_prefix(self):
        local_url = "https://myhost.example.com"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual(None, port)
        self.assertEqual(None, prefix)

    def test_hostname_no_scheme_port_no_prefix(self):
        local_url = "myhost.example.com:8443"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("myhost.example.com", hostname)
        self.assertEqual("8443", port)
        self.assertEqual(None, prefix)

    def test_just_prefix(self):
        local_url = "/myapp"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual(None, hostname)
        self.assertEqual(None, port)
        self.assertEqual("/myapp", prefix)

    def test_short_name(self):
        # could argue anything could be a local hostname, and we should
        # use default port and path. You could also argue it should
        # throw an error, especially if it's not a valid hostname
        local_url = "a"
        (username, password, hostname, port, prefix) = parse_url(
            local_url, default_port=DEFAULT_PORT, default_prefix=DEFAULT_PREFIX
        )
        self.assertEqual("a", hostname)
        self.assertEqual(DEFAULT_PORT, port)
        self.assertEqual(DEFAULT_PREFIX, prefix)

    def test_wrong_scheme(self):
        local_url = "git://git.fedorahosted.org/candlepin.git"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    def test_bad_http_scheme(self):
        # note missing /
        local_url = "https:/myhost.example.com:8443/myapp"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    def test_colon_but_no_port(self):
        local_url = "https://myhost.example.com:/myapp"
        self.assertRaises(ServerUrlParseErrorPort, parse_url, local_url)

    def test_colon_but_no_port_no_scheme(self):
        local_url = "myhost.example.com:/myapp"
        self.assertRaises(ServerUrlParseErrorPort, parse_url, local_url)

    def test_colon_slash_slash_but_nothing_else(self):
        local_url = "http://"
        self.assertRaises(ServerUrlParseErrorJustScheme, parse_url, local_url)

    def test_colon_slash_but_nothing_else(self):
        local_url = "http:/"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    def test_colon_no_slash(self):
        local_url = "http:example.com/foobar"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    # Note: this means if you have a local server named
    # "http", and you like redundant slashes, this actually
    # valid url of http//path/to/something will fail.
    # Don't do that. (or just use a single slash like http/path)
    # But seriously, really?
    def test_no_colon_double_slash(self):
        local_url = "http//example.com/api"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    def test_https_no_colon_double_slash(self):
        local_url = "https//example.com/api"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    # fail at internet
    def test_just_colon_slash(self):
        local_url = "://"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    def test_one_slash(self):
        local_url = "http/example.com"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    def test_host_named_http(self):
        local_url = "http://http/prefix"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("http", hostname)
        self.assertEqual(None, port)
        self.assertEqual("/prefix", prefix)

    def test_one_slash_port_prefix(self):
        local_url = "https/bogaddy:80/candlepin"
        self.assertRaises(ServerUrlParseErrorScheme, parse_url, local_url)

    def test_host_named_http_port_prefix(self):
        local_url = "https://https:8000/prefix"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("https", hostname)
        self.assertEqual("8000", port)
        self.assertEqual("/prefix", prefix)

    def test_host_name_non_numeric_port(self):
        local_url = "https://example.com:https/prefix"
        self.assertRaises(ServerUrlParseErrorPort, parse_url, local_url)


class TestRemoveScheme(unittest.TestCase):
    def test_colon_port(self):
        proxy_url = "proxy.example.com:3128"
        res = remove_scheme(proxy_url)
        self.assertEqual(res, proxy_url)

    def test_http_scheme(self):
        proxy_url = "http://example.com:3128"
        res = remove_scheme(proxy_url)
        self.assertEqual(res, "example.com:3128")

    def test_https_scheme(self):
        proxy_url = "https://example.com:3128"
        res = remove_scheme(proxy_url)
        self.assertEqual(res, "example.com:3128")

    def test_no_port(self):
        proxy_url = "proxy.example.com"
        res = remove_scheme(proxy_url)
        self.assertEqual(res, proxy_url)


class TestHasBadScheme(unittest.TestCase):
    def test_bad(self):
        self.assertTrue(has_bad_scheme("://example.com"))
        self.assertTrue(has_bad_scheme("http/example.com"))
        self.assertTrue(has_bad_scheme("https/example.com"))
        self.assertTrue(has_bad_scheme("https:/example.com"))

    def test_good(self):
        self.assertFalse(has_bad_scheme("http://example.com"))
        self.assertFalse(has_bad_scheme("https://example.com"))


class TestHasGoodScheme(unittest.TestCase):
    def test_good(self):
        self.assertTrue(has_good_scheme("http://example.com"))
        self.assertTrue(has_good_scheme("https://example.com"))

    def test_bad(self):
        self.assertFalse(has_good_scheme("://example.com"))
        self.assertFalse(has_good_scheme("http/example.com"))
        self.assertFalse(has_good_scheme("https/example.com"))
        self.assertFalse(has_good_scheme("https:/example.com"))


class TestParseUrl(unittest.TestCase):
    def test_username_password(self):
        local_url = "http://user:pass@hostname:1111/prefix"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("user", username)
        self.assertEqual("pass", password)
        self.assertEqual("hostname", hostname)
        self.assertEqual("1111", port)
        self.assertEqual("/prefix", prefix)

    def test_no_password(self):
        local_url = "http://user@hostname:1111/prefix"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual("user", username)
        self.assertEqual(None, password)
        self.assertEqual("hostname", hostname)
        self.assertEqual("1111", port)
        self.assertEqual("/prefix", prefix)

    def test_no_userinfo(self):
        local_url = "http://hostname:1111/prefix"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual(None, username)
        self.assertEqual(None, password)
        self.assertEqual("hostname", hostname)
        self.assertEqual("1111", port)
        self.assertEqual("/prefix", prefix)

    def test_no_userinfo_with_at(self):
        local_url = "http://@hostname:1111/prefix"
        (username, password, hostname, port, prefix) = parse_url(local_url)
        self.assertEqual(None, username)
        self.assertEqual(None, password)
        self.assertEqual("hostname", hostname)
        self.assertEqual("1111", port)
        self.assertEqual("/prefix", prefix)


class TestProxyInfo(unittest.TestCase):
    def _gen_env(self, variables):
        """
        Return an environment with everything empty except
        those passed in variables.
        """
        proxy_env = {"HTTPS_PROXY": "", "https_proxy": "", "HTTP_PROXY": "", "http_proxy": ""}
        if variables:
            for (key, value) in list(variables.items()):
                proxy_env[key] = value
        return proxy_env

    def test_https_proxy_info(self):
        with patch.dict("os.environ", self._gen_env({"https_proxy": "https://u:p@host:1111"})):
            proxy_info = get_env_proxy_info()
            self.assertEqual("u", proxy_info["proxy_username"])
            self.assertEqual("p", proxy_info["proxy_password"])
            self.assertEqual("host", proxy_info["proxy_hostname"])
            self.assertEqual(int("1111"), proxy_info["proxy_port"])

    def test_http_proxy_info(self):
        with patch.dict("os.environ", self._gen_env({"http_proxy": "http://u:p@host:2222"})):
            proxy_info = get_env_proxy_info()
            self.assertEqual("u", proxy_info["proxy_username"])
            self.assertEqual("p", proxy_info["proxy_password"])
            self.assertEqual("host", proxy_info["proxy_hostname"])
            self.assertEqual(int("2222"), proxy_info["proxy_port"])

    def test_http_proxy_info_allcaps(self):
        with patch.dict("os.environ", self._gen_env({"HTTP_PROXY": "http://u:p@host:3333"})):
            proxy_info = get_env_proxy_info()
            self.assertEqual("u", proxy_info["proxy_username"])
            self.assertEqual("p", proxy_info["proxy_password"])
            self.assertEqual("host", proxy_info["proxy_hostname"])
            self.assertEqual(int("3333"), proxy_info["proxy_port"])

    def test_https_proxy_info_allcaps(self):
        with patch.dict("os.environ", self._gen_env({"HTTPS_PROXY": "https://u:p@host:4444"})):
            proxy_info = get_env_proxy_info()
            self.assertEqual("u", proxy_info["proxy_username"])
            self.assertEqual("p", proxy_info["proxy_password"])
            self.assertEqual("host", proxy_info["proxy_hostname"])
            self.assertEqual(int("4444"), proxy_info["proxy_port"])

    def test_order(self):
        # should follow the order: HTTPS, https, HTTP, http
        with patch.dict(
            "os.environ",
            self._gen_env(
                {"HTTPS_PROXY": "http://u:p@host:1111", "http_proxy": "http://notme:orme@host:2222"}
            ),
        ):
            proxy_info = get_env_proxy_info()
            self.assertEqual("u", proxy_info["proxy_username"])
            self.assertEqual("p", proxy_info["proxy_password"])
            self.assertEqual("host", proxy_info["proxy_hostname"])
            self.assertEqual(int("1111"), proxy_info["proxy_port"])

    def test_no_port(self):
        with patch.dict("os.environ", self._gen_env({"HTTPS_PROXY": "http://u:p@host"})):
            proxy_info = get_env_proxy_info()
            self.assertEqual("u", proxy_info["proxy_username"])
            self.assertEqual("p", proxy_info["proxy_password"])
            self.assertEqual("host", proxy_info["proxy_hostname"])
            self.assertEqual(3128, proxy_info["proxy_port"])

    def test_no_user_or_password(self):
        with patch.dict("os.environ", self._gen_env({"HTTPS_PROXY": "http://host:1111"})):
            proxy_info = get_env_proxy_info()
            self.assertEqual(None, proxy_info["proxy_username"])
            self.assertEqual(None, proxy_info["proxy_password"])
            self.assertEqual("host", proxy_info["proxy_hostname"])
            self.assertEqual(int("1111"), proxy_info["proxy_port"])


class TestCmdName(unittest.TestCase):
    def test_usr_sbin(self):
        argv = ["/usr/sbin/subscription-manager", "list"]
        self.assertEqual("subscription-manager", cmd_name(argv))

    def test_bin(self):
        argv = ["bin/subscription-manager", "subscribe", "--auto"]
        self.assertEqual("subscription-manager", cmd_name(argv))

    def test_sbin(self):
        argv = ["/sbin/subscription-manager", "list"]
        self.assertEqual("subscription-manager", cmd_name(argv))

    def test_subscription_manager_gui(self):
        argv = ["/sbin/subscription-manager-gui"]
        self.assertEqual("subscription-manager-gui", cmd_name(argv))

    def test_yum(self):
        argv = ["/bin/yum", "install", "zsh"]
        self.assertEqual("yum", cmd_name(argv))

    def test_rhsmcertd_worker(self):
        argv = ["/usr/libexec/rhsmcertd-worker"]
        self.assertEqual("rhsmcertd-worker", cmd_name(argv))

    def test_rhsm_debug(self):
        argv = ["/bin/rhsm-debug"]
        self.assertEqual("rhsm-debug", cmd_name(argv))

    def test_virt_who(self):
        argv = ["/usr/share/virt-who/virtwho.py"]
        self.assertEqual("virtwho.py", cmd_name(argv))


class TestSingletonize(unittest.TestCase):
    def test_class(self):
        """Test that two instances of singleton class are the same object."""

        @singleton
        class Singleton:
            pass

        singleton_1 = Singleton()
        singleton_2 = Singleton()
        self.assertEqual(id(singleton_1), id(singleton_2))
        self.assertEqual(singleton_1, singleton_2)

    def test_subclass(self):
        """Test that two singleton subclass instances are the same object."""

        class Parent:
            pass

        @singleton
        class Child(Parent):
            pass

        child_1 = Child()
        child_2 = Child()
        self.assertEqual(id(child_1), id(child_2))
        self.assertEqual(child_1, child_2)

    def test_subclass_of_singleton(self):
        """Test that singleton's subclass is also singleton."""

        @singleton
        class Parent:
            pass

        class Child(Parent):
            pass

        parent = Parent()
        child_1 = Child()
        child_2 = Child()

        self.assertNotEqual(parent, child_1)
        self.assertNotEqual(id(parent), id(child_1))
        self.assertEqual(child_1, child_2)
        self.assertEqual(id(child_1), id(child_2))

        self.assertEqual(type(child_1._instance), Child)
        self.assertEqual(type(child_2._instance), Child)

    def test_class_with_init(self):
        """Test that two instances of singleton share their object attributes."""

        @singleton
        class Singleton:
            def __init__(self, value: int):
                self.value = value

        singleton_1 = Singleton(1)
        singleton_2 = Singleton(2)
        self.assertEqual(singleton_1.value, 2)
        self.assertEqual(singleton_2.value, 2)

    def test_shared_attribute(self):
        """Test that two instances of singleton share their class attributes."""

        @singleton
        class Singleton:
            value: int = 0

            def __init__(self):
                self.value += 1

        singleton_1 = Singleton()
        singleton_2 = Singleton()
        self.assertEqual(singleton_1.value, 2)
        self.assertEqual(singleton_2.value, 2)

    def test_custom_new(self):
        """Test that singleton class works with custom __new__ methods."""

        @singleton
        class Singleton:
            value: int = 0

            def __new__(cls, value: int):
                cls.value = value
                return object.__new__(cls)

        singleton_1 = Singleton(1)
        singleton_2 = Singleton(2)

        self.assertEqual(singleton_1.value, 1)
        self.assertEqual(singleton_2.value, 1)


class TestCallOnce(unittest.TestCase):
    def test_basic(self):
        """Test that function with no parent class is called just once."""

        @call_once
        def add(a: int, b: int) -> int:
            return a + b

        result_1 = add(1, 2)
        self.assertEqual(result_1, 3)
        result_2 = add(3, 4)
        self.assertIsNone(result_2)

    def test_init(self):
        """Test that __init__ is called just once."""

        class Foo:
            value: int = 0

            @call_once
            def __init__(self, value: int):
                self.value = value

        foo1 = Foo(1)
        self.assertEqual(foo1.value, 1)
        foo2 = Foo(2)
        self.assertEqual(foo2.value, 0)

    def test_reset(self):
        """Test that call_once-decorated function can be reset."""

        class Foo:
            value: int = 0

            @call_once
            def __init__(self, value: int):
                self.value = value

        foo1 = Foo(1)
        self.assertEqual(foo1.value, 1)
        foo2 = Foo(2)
        self.assertEqual(foo2.value, 0)

        Foo.__init__._reset()

        foo3 = Foo(3)
        self.assertEqual(foo3.value, 3)
        foo4 = Foo(4)
        self.assertEqual(foo4.value, 0)

    def test_classmethod(self):
        """Test that classmethod is called just once."""

        class Foo:
            value: int = 0

            @classmethod
            @call_once
            def increase(cls):
                cls.value += 1

        foo = Foo()
        foo.increase()
        self.assertEqual(foo.value, 1)
        foo.increase()
        self.assertEqual(foo.value, 1)

    def test_staticmethod(self):
        """Test that staticmethod is called just once."""

        class Foo:
            value: int = 0

            @staticmethod
            @call_once
            def increase():
                Foo.value += 1

        foo = Foo()
        foo.increase()
        self.assertEqual(foo.value, 1)
        foo.increase()
        self.assertEqual(foo.value, 1)

    def test_callonce_inheritance(self):
        """Test that @call_once is inherited."""

        class Parent:
            @call_once
            def __init__(self, value: int):
                self.value = value

            @property
            def value(self):
                return self._value

            @value.setter
            def value(self, value: int):
                self._value = value

        class Child(Parent):
            def __init__(self, value: int):
                super().__init__(value)

        child_1 = Child(1)
        child_2 = Child(2)
        self.assertEqual(child_1.value, 1)
        with self.assertRaises(AttributeError) as exc_info:
            getattr(child_2, "value")
        self.assertEqual(str(exc_info.exception), "'Child' object has no attribute '_value'")


class TestLock(unittest.TestCase):
    """Test @lock decorator."""

    def test_lock_unlock(self):
        """Test lock(), unlock() and 'locked' property."""

        @lock
        class TestLock:
            pass

        test_lock = TestLock()

        self.assertEqual(test_lock.locked, False)
        test_lock.lock()
        self.assertEqual(test_lock.locked, True)
        test_lock.unlock()
        self.assertEqual(test_lock.locked, False)

    def test_unlocking_of_not_locked_lock(self):
        """Test that exception is not raised when we try to unlock not locked lock."""

        @lock
        class TestLock:
            pass

        test_lock = TestLock()

        self.assertEqual(test_lock.locked, False)
        test_lock.unlock()
        self.assertEqual(test_lock.locked, False)

    def test_locking_of_locked_lock(self):
        """Test that lock can be locked multiple times."""

        @lock
        class TestLock:
            pass

        test_lock = TestLock()

        test_lock.lock()
        self.assertEqual(test_lock.locked, True)
        test_lock.lock()
        self.assertEqual(test_lock.locked, True)

    def test_context_manager(self):
        """Test lock as context manager."""

        @lock
        class TestLock:
            pass

        with TestLock() as test_lock:
            self.assertEqual(test_lock.locked, True)
        self.assertEqual(test_lock.locked, False)

    def test_cannot_set_attribute(self):
        """Test that it's not possible to assign values to 'locked' attribute."""

        @lock
        class TestLock:
            pass

        test_lock = TestLock()

        with self.assertRaises(AttributeError):
            test_lock.locked = True


class TestComplexSingleton(unittest.TestCase):
    """Test combinations of @singletonize, @call_once and @lock."""

    def test_init(self):
        """Test that __init__ of singleton is called just once."""

        @singleton
        class Singleton:
            value = 0

            @call_once
            def __init__(self):
                self.value += 1

        singleton_1 = Singleton()
        singleton_2 = Singleton()
        self.assertEqual(singleton_1.value, 1)
        self.assertEqual(singleton_2.value, 1)

    def test_callonce_inheritance(self):
        """Test that @call_once is inherited in singletons."""

        @singleton
        class Parent:
            @call_once
            def __init__(self, value: int):
                self.value = value

            @property
            def value(self):
                return self._value

            @value.setter
            def value(self, value: int):
                self._value = value

        class Child(Parent):
            def __init__(self, value: int):
                super().__init__(value)

        child_1 = Child(1)
        self.assertEqual(child_1.value, 1)
        child_2 = Child(2)
        self.assertEqual(child_1, child_2)
        self.assertEqual(child_2.value, 1)

    def test_exception_does_not_prevent_lock_initialization(self):
        """Test that function is marked as called even if it raises an exception."""

        @lock
        @singleton
        class TestLock:
            @call_once
            def __init__(self):
                raise Exception("I am designed to fail exactly once.")

        with self.assertRaises(Exception):
            TestLock()

        self.assertIsNotNone(TestLock._instance)
        test_lock_1_id = id(TestLock._instance)

        test_lock_2 = TestLock()
        self.assertEqual(id(test_lock_2), test_lock_1_id)

    def test_unlock_with_multiple_instances(self):
        """Test that singleton lock is unlocked for all instances."""

        @lock
        @singleton
        class TestLock:
            pass

        test_lock_1 = TestLock()
        test_lock_2 = TestLock()

        self.assertEqual(test_lock_1, test_lock_2)

        test_lock_1.lock()
        self.assertEqual(test_lock_1.locked, True)
        self.assertEqual(test_lock_2.locked, True)

        test_lock_2.lock()
        self.assertEqual(test_lock_1.locked, True)
        self.assertEqual(test_lock_2.locked, True)

        test_lock_1.unlock()
        self.assertEqual(test_lock_1.locked, False)
        self.assertEqual(test_lock_2.locked, False)

    def test_threading(self):
        """Test that singletons are one object in two threads."""

        @singleton
        class Singleton:
            @call_once
            def __init__(self, foo: str, bar: str):
                self.foo = foo
                self.bar = bar

        def thread_function(foo, bar) -> None:
            Singleton(foo, bar)
            time.sleep(0.3)

        thread_1 = threading.Thread(target=thread_function, args=("foo", "bar"))
        thread_2 = threading.Thread(target=thread_function, args=("FOO", "BAR"))

        thread_1.start()
        time.sleep(0.1)

        thread_2.start()
        time.sleep(0.1)

        thread_1.join(timeout=0.5)
        thread_2.join(timeout=0.5)

        singleton_1 = Singleton()
        self.assertEqual(singleton_1.foo, "foo")
        self.assertEqual(singleton_1.bar, "bar")

    def test_lock_threading(self):
        """Test that singleton locks are one object in two threads."""

        @lock
        @singleton
        class TestLock:
            foo: str
            bar: str

        def thread_function(foo, bar) -> None:
            with TestLock() as test_lock:
                test_lock.foo = foo
                test_lock.bar = bar
                time.sleep(0.3)

        thread_1 = threading.Thread(target=thread_function, args=("foo", "bar"))
        thread_2 = threading.Thread(target=thread_function, args=("FOO", "BAR"))

        thread_1.start()
        time.sleep(0.1)

        thread_2.start()
        time.sleep(0.1)

        thread_1.join(timeout=0.5)
        thread_2.join(timeout=0.5)

        lock_1 = TestLock()
        self.assertEqual(lock_1.foo, "FOO")
        self.assertEqual(lock_1.bar, "BAR")
